import csv
import os
import sys
import threading
import traceback

from flask import g

import requests

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, or_

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import (
    Contract,
    Era,
    MtcParticipant,
    Party,
    ReportRun,
    Session,
    Source,
    Supply,
    User,
)
from chellow.utils import req_bool
from chellow.views import chellow_redirect


FNAME = "ecoes_comparison"


def _parse_date(date_str):
    if len(date_str) > 0:
        day, month, year = date_str.split("/")
        return f"{year}-{month}-{day} 00:00"
    else:
        return date_str


def content(user_id, show_ignored, report_run_id):
    f = report_run = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            running_name, finished_name = chellow.dloads.make_names(
                f"{FNAME}.csv", user
            )
            f = open(running_name, mode="w", newline="")
            report_run = ReportRun.get_by_id(sess, report_run_id)

            props = Contract.get_non_core_by_name(
                sess, "configuration"
            ).make_properties()

            ECOES_KEY = "ecoes"
            try:
                ecoes_props = props[ECOES_KEY]
            except KeyError:
                raise BadRequest(
                    f"The property {ECOES_KEY} cannot be found in the configuration "
                    f"properties."
                )

            for key in (
                "user_name",
                "password",
                "prefix",
                "exclude_mpan_cores",
                "ignore_mpan_cores_msn",
            ):
                try:
                    ecoes_props[key]
                except KeyError:
                    raise BadRequest(
                        f"The property {key} cannot be found in the 'ecoes' section of "
                        f"the configuration properties."
                    )

            exclude_mpan_cores = ecoes_props["exclude_mpan_cores"]
            ignore_mpan_cores_msn = ecoes_props["ignore_mpan_cores_msn"]
            url_prefix = ecoes_props["prefix"]

            proxies = props.get("proxies", {})
            s = requests.Session()
            s.verify = False
            r = s.get(url_prefix, proxies=proxies)
            data = {
                "Username": ecoes_props["user_name"],
                "Password": ecoes_props["password"],
            }
            login_j = s.post(url_prefix, data=data, allow_redirects=False).json()
            if not login_j["Success"]:
                raise BadRequest(f"Login to ECOES failed: {login_j['Messages']}")
            elif "RedirectUrl" in login_j and "SetPassword" in login_j["RedirectUrl"]:
                raise BadRequest(
                    "Login to ECOES failed, it looks like the password needs to be "
                    "changed"
                )

            r = s.get(
                f"{url_prefix}PortfolioAccess/ExportPortfolioMPANs?fileType=csv",
                proxies=proxies,
            )

            _process(
                sess,
                r.text.splitlines(True),
                exclude_mpan_cores,
                ignore_mpan_cores_msn,
                f,
                show_ignored,
                report_run,
            )
            report_run.update("finished")
            sess.commit()

    except BadRequest as e:
        msg = e.description
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["problem"], {"problem": msg}, {})
            sess.commit()
        if f is not None:
            f.write(msg)
        raise e
    except BaseException as e:
        msg = traceback.format_exc()
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["problem"], {"problem": msg}, {})
            sess.commit()
        sys.stderr.write(msg)
        if f is not None:
            f.write(msg)
        raise e
    finally:
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _process(
    sess,
    ecoes_lines,
    exclude_mpan_cores,
    ignore_mpan_cores_msn,
    f,
    show_ignored,
    report_run,
):
    writer = csv.writer(f, lineterminator="\n")

    mpans = []

    for exp in (Era.imp_mpan_core, Era.exp_mpan_core):
        for (v,) in sess.execute(
            select(exp)
            .join(Supply)
            .join(Source)
            .join(Supply.dno)
            .filter(
                Party.dno_code.notin_(("88", "99")),
                Era.finish_date == null(),
                Source.code != "3rd-party",
                exp.notin_(exclude_mpan_cores),
                exp != null(),
            )
            .distinct()
            .order_by(exp)
        ):
            mpans.append(v)

    ecoes_mpans = {}

    parser = iter(csv.reader(ecoes_lines))
    next(parser)  # Skip titles

    for values in parser:
        ecoes_titles = [
            "mpan-core",
            "address-line-1",
            "address-line-2",
            "address-line-3",
            "address-line-4",
            "address-line-5",
            "address-line-6",
            "address-line-7",
            "address-line-8",
            "address-line-9",
            "post-code",
            "supplier",
            "registration-from",
            "mtc",
            "mtc-date",
            "llfc",
            "llfc-from",
            "pc",
            "ssc",
            "measurement-class",
            "energisation-status",
            "da",
            "dc",
            "mop",
            "mop-appoint-date",
            "gsp-group",
            "gsp-effective-from",
            "dno",
            "msn",
            "meter-install-date",
            "meter-type",
            "map-id",
        ]

        ecoes_row = dict(zip(ecoes_titles, map(str.strip, values)))
        mpan_core = ecoes_row["mpan-core"]
        mpan_spaces = " ".join(
            (
                mpan_core[:2],
                mpan_core[2:6],
                mpan_core[6:10],
                mpan_core[-3:],
            )
        )
        if mpan_spaces in exclude_mpan_cores:
            continue

        ecoes_row["mpan_spaces"] = mpan_spaces
        if mpan_spaces in ecoes_mpans:
            prev_row = ecoes_mpans[mpan_spaces]
            prev_row["meter_count"] += 1
        else:
            ecoes_row["meter_count"] = 1
            ecoes_mpans[mpan_spaces] = ecoes_row

    titles = (
        "mpan_core",
        "mpan_core_no_spaces",
        "ecoes_pc",
        "chellow_pc",
        "ecoes_mtc",
        "chellow_mtc",
        "ecoes_llfc",
        "chellow_llfc",
        "ecoes_ssc",
        "chellow_ssc",
        "ecoes_es",
        "chellow_es",
        "ecoes_supplier",
        "chellow_supplier",
        "chellow_supplier_contract_name",
        "ecoes_dc",
        "chellow_dc",
        "ecoes_mop",
        "chellow_mop",
        "ecoes_gsp_group",
        "chellow_gsp_group",
        "ecoes_msn",
        "chellow_msn",
        "ecoes_msn_install_date",
        "ecoes_meter_type",
        "chellow_meter_type",
        "ignored",
        "problem",
    )
    writer.writerow(titles)

    for mpan_spaces, ecoes in sorted(ecoes_mpans.items()):
        problem = ""
        chellow_pc = ""
        chellow_mtc = ""
        chellow_llfc = ""
        chellow_ssc = ""
        chellow_es = ""
        chellow_supplier = ""
        chellow_supplier_contract_name = ""
        chellow_supplier_contract_id = None
        chellow_dc = ""
        chellow_mop = ""
        chellow_gsp_group = ""
        chellow_msn = ""
        chellow_meter_type = ""
        chellow_supply_id = None
        chellow_era_id = None
        ignore = True
        diffs = []

        try:
            ecoes_es = ecoes["energisation-status"]
        except KeyError as e:
            raise e

        ecoes_disconnected = ecoes_es == ""
        current_chell = mpan_spaces in mpans

        if ecoes["meter_count"] > 1:
            problem += (
                f"There are {ecoes['meter_count']} meters associated with this MPAN "
                f"core in ECOES, but Chellow only supports one meter per supply at the "
                f"moment. If there really should be multiple meters for this supply, "
                f"let me know and I'll add support for it in Chellow."
            )
            ignore = False

        if ecoes_disconnected and current_chell:
            problem += "Disconnected in ECOES, but current in Chellow. "
            ignore = False

        elif not ecoes_disconnected and not current_chell:
            era = sess.execute(
                select(Era)
                .filter(
                    or_(
                        Era.imp_mpan_core == mpan_spaces,
                        Era.exp_mpan_core == mpan_spaces,
                    ),
                )
                .order_by(Era.start_date.desc())
                .limit(1)
            ).scalar_one_or_none()
            if era is None:
                problem += f"In ECOES (as {ecoes_es}) but not in Chellow. "
            else:
                problem += f"In ECOES (as {ecoes_es}) but disconnected in Chellow. "
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                else:
                    supplier_contract = era.exp_supplier_contract
                chellow_supplier_contract_name = supplier_contract.name
                chellow_supplier_contract_id = supplier_contract.id
                chellow_supply_id = era.supply.id
                chellow_era_id = era.id
            ignore = False

        if current_chell:
            mpans.remove(mpan_spaces)
            era = sess.execute(
                select(Era)
                .filter(
                    Era.finish_date == null(),
                    or_(
                        Era.imp_mpan_core == mpan_spaces,
                        Era.exp_mpan_core == mpan_spaces,
                    ),
                )
                .options(
                    joinedload(Era.supply).joinedload(Supply.gsp_group),
                    joinedload(Era.mop_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.dc_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.imp_supplier_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.exp_supplier_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.pc),
                    joinedload(Era.imp_llfc),
                    joinedload(Era.exp_llfc),
                    joinedload(Era.mtc_participant).joinedload(
                        MtcParticipant.meter_type
                    ),
                    joinedload(Era.ssc),
                    joinedload(Era.energisation_status),
                    joinedload(Era.channels),
                )
            ).scalar()
            chellow_supply_id = era.supply.id
            chellow_era_id = era.id
            chellow_es = era.energisation_status.code
            if ecoes_es != chellow_es:
                problem += "The energisation statuses don't match. "
                ignore = False
                diffs.append("es")

            if not (ecoes_es == "D" and chellow_es == "D"):
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                    llfc = era.imp_llfc
                else:
                    supplier_contract = era.exp_supplier_contract
                    llfc = era.exp_llfc

                chellow_pc = era.pc.code
                try:
                    if int(ecoes["pc"]) != int(chellow_pc):
                        problem += "The PCs don't match. "
                        ignore = False
                        diffs.append("pc")
                except ValueError:
                    problem += "Can't parse the PC. "
                    ignore = False

                chellow_mtc = era.mtc_participant.mtc.code
                try:
                    if int(ecoes["mtc"]) != int(chellow_mtc):
                        problem += "The MTCs don't match. "
                        ignore = False
                        diffs.append("mtc")
                except ValueError:
                    problem += "Can't parse the MTC. "
                    ignore = False

                chellow_llfc = llfc.code
                if ecoes["llfc"].zfill(3) != chellow_llfc:
                    problem += "The LLFCs don't match. "
                    ignore = False
                    diffs.append("llfc")

                chellow_ssc = era.ssc
                if chellow_ssc is None:
                    chellow_ssc = ""
                    chellow_ssc_int = None
                else:
                    chellow_ssc = chellow_ssc.code
                    chellow_ssc_int = int(chellow_ssc)

                if len(ecoes["ssc"]) > 0:
                    ecoes_ssc_int = int(ecoes["ssc"])
                else:
                    ecoes_ssc_int = None

                if ecoes_ssc_int != chellow_ssc_int and not (
                    ecoes_ssc_int is None and chellow_ssc_int is None
                ):
                    problem += "The SSCs don't match. "
                    ignore = False
                    diffs.append("ssc")

                chellow_supplier = supplier_contract.party.participant.code
                chellow_supplier_contract_name = supplier_contract.name
                chellow_supplier_contract_id = supplier_contract.id
                if chellow_supplier != ecoes["supplier"]:
                    problem += "The supplier codes don't match. "
                    ignore = False
                    diffs.append("supplier")

                dc_contract = era.dc_contract
                if dc_contract is None:
                    chellow_dc = ""
                else:
                    chellow_dc = dc_contract.party.participant.code

                if chellow_dc != ecoes["dc"]:
                    problem += "The DC codes don't match. "
                    ignore = False
                    diffs.append("dc")

                mop_contract = era.mop_contract
                if mop_contract is None:
                    chellow_mop = ""
                else:
                    chellow_mop = mop_contract.party.participant.code

                if chellow_mop != ecoes["mop"]:
                    problem += "The MOP codes don't match. "
                    ignore = False
                    diffs.append("mop")

                chellow_gsp_group = era.supply.gsp_group.code
                if chellow_gsp_group != ecoes["gsp-group"]:
                    problem += "The GSP group codes don't match. "
                    ignore = False
                    diffs.append("gsp_group")

                chellow_msn = era.msn
                if chellow_msn is None:
                    chellow_msn = ""

                if chellow_msn != ecoes["msn"]:
                    problem += "The meter serial numbers don't match. "
                    diffs.append("msn")
                    if mpan_spaces not in ignore_mpan_cores_msn:
                        ignore = False
                elif mpan_spaces in ignore_mpan_cores_msn:
                    problem += (
                        "This MPAN core is in mpan_cores_ignore and yet the meter "
                        "serial numbers do match. "
                    )

                chellow_meter_type = _meter_type(era)

                if chellow_meter_type != ecoes["meter-type"]:
                    problem += "The DTC meter types don't match."
                    ignore = False
                    diffs.append("meter_type")

        if len(problem) > 0 and not (not show_ignored and ignore):
            values = {
                "mpan_core": mpan_spaces,
                "mpan_core_no_spaces": ecoes["mpan-core"],
                "ecoes_pc": ecoes["pc"],
                "chellow_pc": chellow_pc,
                "ecoes_mtc": ecoes["mtc"],
                "chellow_mtc": chellow_mtc,
                "chellow_mtc_date": _parse_date(ecoes["mtc-date"]),
                "ecoes_llfc": ecoes["llfc"],
                "ecoes_llfc_from": _parse_date(ecoes["llfc-from"]),
                "chellow_llfc": chellow_llfc,
                "ecoes_ssc": ecoes["ssc"],
                "chellow_ssc": chellow_ssc,
                "ecoes_es": ecoes["energisation-status"],
                "chellow_es": chellow_es,
                "ecoes_supplier": ecoes["supplier"],
                "ecoes_supplier_registration_from": _parse_date(
                    ecoes["registration-from"]
                ),
                "chellow_supplier": chellow_supplier,
                "chellow_supplier_contract_name": chellow_supplier_contract_name,
                "ecoes_dc": ecoes["dc"],
                "chellow_dc": chellow_dc,
                "ecoes_mop": ecoes["mop"],
                "ecoes_mop_appoint_date": _parse_date(ecoes["mop-appoint-date"]),
                "chellow_mop": chellow_mop,
                "ecoes_gsp_group": ecoes["gsp-group"],
                "ecoes_gsp_effective_from": _parse_date(ecoes["gsp-effective-from"]),
                "chellow_gsp_group": chellow_gsp_group,
                "ecoes_msn": ecoes["msn"],
                "chellow_msn": chellow_msn,
                "ecoes_msn_install_date": _parse_date(ecoes["meter-install-date"]),
                "ecoes_meter_type": ecoes["meter-type"],
                "chellow_meter_type": chellow_meter_type,
                "ignored": ignore,
                "problem": problem,
            }
            writer.writerow(values[t] for t in titles)
            values["chellow_supplier_contract_id"] = chellow_supplier_contract_id
            values["chellow_supply_id"] = chellow_supply_id
            values["diffs"] = diffs
            values["chellow_era_id"] = chellow_era_id
            report_run.insert_row(sess, "", titles, values, {})
            sess.commit()
        sess.expunge_all()

    for mpan_core in mpans:
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        era = supply.find_era_at(sess, None)
        if era.imp_mpan_core == mpan_core:
            supplier_contract = era.imp_supplier_contract
            llfc = era.imp_llfc
        else:
            supplier_contract = era.exp_supplier_contract
            llfc = era.exp_llfc

        ssc = "" if era.ssc is None else era.ssc.code

        es = era.energisation_status.code

        dc_contract = era.dc_contract
        dc = "" if dc_contract is None else dc_contract.party.participant.code

        mop_contract = era.mop_contract
        mop = "" if mop_contract is None else mop_contract.party.participant.code

        msn = "" if era.msn is None else era.msn

        meter_type = _meter_type(era)

        values = {
            "mpan_core": mpan_core,
            "mpan_core_no_spaces": mpan_core.replace(" ", ""),
            "ecoes_pc": "",
            "chellow_pc": era.pc.code,
            "ecoes_mtc": "",
            "chellow_mtc": era.mtc_participant.mtc.code,
            "ecoes_llfc": "",
            "chellow_llfc": llfc.code,
            "ecoes_ssc": "",
            "chellow_ssc": ssc,
            "ecoes_es": "",
            "chellow_es": es,
            "ecoes_supplier": "",
            "chellow_supplier": supplier_contract.party.participant.code,
            "chellow_supplier_contract_name": supplier_contract.name,
            "ecoes_dc": "",
            "chellow_dc": dc,
            "ecoes_mop": "",
            "chellow_mop": mop,
            "ecoes_gsp_group": "",
            "chellow_gsp_group": supply.gsp_group.code,
            "ecoes_msn": "",
            "chellow_msn": msn,
            "ecoes_msn_install_date": "",
            "ecoes_meter_type": "",
            "chellow_meter_type": meter_type,
            "ignored": False,
            "problem": "In Chellow, but not in ECOES.",
        }
        writer.writerow(values[t] for t in titles)
        values["chellow_supplier_contract_id"] = supplier_contract.id
        values["chellow_supply_id"] = era.supply.id
        values["diffs"] = []
        values["chellow_era_id"] = era.id
        report_run.insert_row(sess, "", titles, values, {})


def _meter_type(era):
    props = era.props
    try:
        return props["meter_type"]
    except KeyError:
        if era.pc.code == "00":
            return "H"
        elif len(era.channels) > 0:
            return "RCAMR"
        elif era.mtc_participant.meter_type.code in ["UM", "PH"]:
            return ""
        else:
            return "N"


def do_get(sess):
    show_ignored = req_bool("show_ignored")
    report_run = ReportRun.insert(
        sess,
        FNAME,
        g.user,
        FNAME,
        {},
    )
    sess.commit()
    args = g.user.id, show_ignored, report_run.id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect(f"/report_runs/{report_run.id}", 303)



from decimal import Decimal

import pytest

from werkzeug.exceptions import BadRequest

import xlrd.sheet

from chellow.e.bill_parsers.engie_xls import _bd_add, _make_raw_bills, _parse_row
from chellow.utils import utc_datetime


def test_parse_row(mocker):
    row = []
    for val in [
        "Power Comany Ltd.",
        "Bill Paja",
        "556",
        "BILL PAJA",
        "883",
        "1 True Way",
        "672770",
        43555.00,
        43555.00,
        "",
        "2019-03-01 - 2019-03-31",
        "",
        "",
        "",
        "Draft",
        "",
        "Product",
        "Hand Held Read -",
        43555.00,
        43555.00,
        "",
        "2299999999929",
        "",
        "",
        "",
        "785",
        "GBP",
        "INV",
        "",
        "",
    ]:
        cell = mocker.Mock(spec=xlrd.sheet.Cell)
        cell.value = val
        row.append(cell)

    row_index = 2
    datemode = 0
    title_row = ["To Date"]

    bill = _parse_row(row, row_index, datemode, title_row)
    assert bill["finish_date"] == utc_datetime(2019, 3, 31, 22, 30)


def test_bd_add():
    el_name = "duos_red"
    bd = {el_name: 0}
    val = None

    with pytest.raises(BadRequest):
        _bd_add(bd, el_name, val)


def test_bill_parser_engie_xls_billed_kwh(mocker):
    """
    Blank units should still give billed kWh
    """
    row_vals = [
        "Mistral Wind Power Ltd.",
        "Bill Paja",
        "886572998",
        "Bill Paja",
        "869987122",
        "BA1 5TT",
        "99708221",
        42627,
        42694,
        42629,
        "2016-08-01 - 2016-08-31",
        "458699",
        "Standard Deluxe",
        "Beautiful Breeze",
        "Accepted",
        "Pass Thru - UK Electricity Cost Component",
        "Product",
        "Renewables Obligation (RO)",
        42583,
        42613,
        "Commercial UK Energy VAT",
        "2298132107763",
        27997.33,
        "",
        0.0971123676,
        "9224",
        "GBP",
        "INV",
        "Renewables Obligation (RO)",
        "",
    ]
    row = []
    for row_val in row_vals:
        v = mocker.Mock()
        v.value = row_val
        row.append(v)

    datemode = mocker.Mock()
    bill = _parse_row(row, 1, datemode, [])
    assert bill["kwh"] == Decimal("27997.33")


def test_make_raw_bills_vat(mocker):
    row_vals = [
        "Mistral Wind Power Ltd.",
        "Bill Paja",
        "886572998",
        "Bill Paja",
        "869987122",
        "BA1 5TT",
        "99708221",
        42627,
        42694,
        42629,
        "2016-08-01 - 2016-08-31",
        "",
        "",
        "",
        "Accepted",
        "",
        "Sales Tax",
        "Standard VAT@20%",
        "",
        "",
        "",
        "2298132107763",
        "",
        "",
        "",
        "9224",
        "GBP",
        "INV",
        "",
        "",
    ]
    row = []
    for row_val in row_vals:
        v = mocker.Mock()
        v.value = row_val
        row.append(v)

    datemode = mocker.Mock()
    sheet = mocker.Mock()
    sheet.nrows = 2
    sheet.row = mocker.Mock(return_value=row)
    bills = _make_raw_bills(sheet, datemode)
    expected_bills = [
        {
            "bill_type_code": "N",
            "kwh": Decimal("0"),
            "vat": Decimal("9224.00"),
            "net": Decimal("0.00"),
            "reads": [],
            "breakdown": {
                "raw_lines": [str(row)],
                "vat_percentage": Decimal("20"),
            },
            "account": "22 9813 2107 763",
            "issue_date": None,
            "start_date": utc_datetime(2016, 7, 31, 23, 0),
            "finish_date": utc_datetime(2016, 8, 31, 22, 30),
            "mpan_core": "22 9813 2107 763",
            "reference": "99708221_2",
            "gross": Decimal("9224.00"),
        }
    ]
    assert bills == expected_bills



import atexit
import ftplib
import importlib
import os
import tempfile
import threading
import traceback
from collections import defaultdict, deque
from datetime import timedelta as Timedelta
from io import TextIOWrapper

import jinja2

import paramiko

import pysftp

import requests

from sqlalchemy import null, or_

from werkzeug.exceptions import BadRequest

from chellow.models import Contract, Era, HhDatum, MarketRole, Session
from chellow.utils import (
    HH,
    hh_format,
    hh_max,
    hh_min,
    utc_datetime,
    utc_datetime_now,
)


processes = defaultdict(list)
tasks = {}

extensions = [".df2", ".simple.csv", ".bg.csv"]


class HhDataImportProcess(threading.Thread):
    def __init__(self, dc_contract_id, process_id, istream, file_name, file_size):
        super().__init__(name=f"HH Manual Import: contract {dc_contract_id}")

        self.messages = []
        self.istream = istream
        self.dc_contract_id = dc_contract_id
        self.id = process_id
        if file_size == 0:
            raise BadRequest("File has zero length")

        file_name = file_name.lower()
        self.conv_ext = [ext for ext in extensions if file_name.endswith(ext)]
        if len(self.conv_ext) == 0:
            raise BadRequest(
                f"The extension of the filename '{file_name}' is not one of the "
                f"recognized extensions; {extensions}"
            )
        self.converter = None

    def run(self):
        with Session() as sess:
            try:
                contract = Contract.get_dc_by_id(sess, self.dc_contract_id)
                sess.rollback()
                properties = contract.make_properties()
                mpan_map = properties.get("mpan_map", {})
                mod_name = self.conv_ext[0][1:].replace(".", "_")
                parser_module = importlib.import_module(
                    f"chellow.e.hh_parser_{mod_name}"
                )
                self.converter = parser_module.create_parser(self.istream, mpan_map)
                sess.rollback()
                HhDatum.insert(sess, self.converter, contract)
                sess.commit()
            except BadRequest as e:
                self.messages.append(e.description)
            except BaseException:
                self.messages.append(f"Outer problem {traceback.format_exc()}")

    def get_status(self):
        return (
            "No converter." if self.converter is None else self.converter.get_status()
        )


def get_hh_import_processes(contract_id):
    return processes[contract_id]


def start_hh_import_process(dc_contract_id, istream, file_name, file_size):
    contract_processes = get_hh_import_processes(dc_contract_id)
    id = len(contract_processes)
    process = HhDataImportProcess(dc_contract_id, id, istream, file_name, file_size)
    contract_processes.append(process)
    process.start()
    return process


lock = threading.RLock()


class HhImportTask(threading.Thread):
    def __init__(self, contract_id):
        super().__init__(name=f"HH Automatic Import: contract {contract_id}")
        self.messages = deque()
        self.contract_id = contract_id
        self.importer = None
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.is_error = False
        self.wait_seconds = 30 * 60

    def stop(self):
        self.stopped.set()
        self.going.set()

    def go(self):
        self.going.set()

    def is_locked(self):
        if lock.acquire(False):
            lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )
        if len(self.messages) > 100:
            self.messages.pop()

    def get_status(self):
        return (
            "No importer." if self.importer is None else str(self.importer.get_status())
        )

    def import_file(self, sess):
        found_new = False

        try:
            contract = Contract.get_dc_by_id(sess, self.contract_id)
            properties = contract.make_properties()
            enabled = properties["enabled"]
            if enabled:
                protocol = properties["protocol"]
                self.log("Protocol is " + protocol)
                if protocol == "ftp":
                    found_new = self.ftp_handler(sess, properties, contract)
                elif protocol == "sftp":
                    found_new = self.sftp_handler(sess, properties, contract)
                elif protocol == "https":
                    found_new = https_handler(sess, self.log, properties, contract)
                else:
                    self.log(f"Protocol '{protocol}' not recognized.")
                self.wait_seconds = properties.get("check_minutes", 30) * 60
            else:
                self.log(
                    "Importer is disabled. To enable it, set the 'enabled' property "
                    "to 'True'."
                )
            self.is_error = False
        except BadRequest as e:
            self.log(f"Problem {e}")
            sess.rollback()
            self.is_error = True
        except Exception:
            self.log(f"Unknown Exception {traceback.format_exc()}")
            sess.rollback()
            self.is_error = True
        return found_new

    def import_now(self):
        if lock.acquire(False):
            with Session() as sess:
                try:
                    while self.import_file(sess):
                        pass
                except Exception:
                    self.log(f"Outer Exception {traceback.format_exc()}")
                    self.is_error = True
                finally:
                    lock.release()

    def run(self):
        while not self.stopped.isSet():
            self.import_now()
            timeout = None if self.is_error else self.wait_seconds
            self.going.wait(timeout=timeout)
            self.going.clear()

    def ftp_handler(self, sess, properties, contract):
        host_name = properties["hostname"]
        user_name = properties["username"]
        password = properties["password"]
        try:
            port = properties["port"]
        except KeyError:
            port = None
        file_type = properties["file_type"]
        directories = properties["directories"]
        state = contract.make_state()

        try:
            last_import_keys = state["last_import_keys"]
        except KeyError:
            last_import_keys = {}
            state["last_import_keys"] = last_import_keys

        sess.rollback()
        self.log(f"Connecting to ftp server at {host_name}:{port}.")
        ftp = ftplib.FTP()
        if port is None:
            ftp.connect(host=host_name, timeout=120)
        else:
            ftp.connect(host=host_name, port=port, timeout=120)

        ftp.login(user_name, password)
        home_path = ftp.pwd()

        fl = None

        for directory in directories:
            self.log("Checking the directory '" + directory + "'.")
            try:
                last_import_key = last_import_keys[directory]
            except KeyError:
                last_import_key = last_import_keys[directory] = ""

            dir_path = home_path + "/" + directory
            ftp.cwd(dir_path)
            files = []
            for fname in ftp.nlst():
                fpath = dir_path + "/" + fname
                try:
                    ftp.cwd(fpath)
                    continue  # directory
                except ftplib.error_perm:
                    pass

                key = ftp.sendcmd("MDTM " + fpath).split()[1] + "_" + fname
                if key > last_import_key:
                    files.append((key, fpath))

            if len(files) > 0:
                fl = sorted(files)[0]
                last_import_keys[directory] = fl[0]
                break

        if fl is None:
            self.log("No new files found.")
            ftp.quit()
            self.log("Logged out.")
            return False
        else:
            key, fpath = fl
            self.log(f"Attempting to download {fpath} with key {key}.")
            f = tempfile.TemporaryFile()
            ftp.retrbinary("RETR " + fpath, f.write)
            self.log("File downloaded successfully.")
            ftp.quit()
            self.log("Logged out.")

            f.seek(0, os.SEEK_END)
            fsize = f.tell()
            f.seek(0)
            self.log(f"File size is {fsize} bytes.")
            self.log(f"Treating files as type {file_type}")
            self.importer = HhDataImportProcess(
                self.contract_id, 0, TextIOWrapper(f, "utf8"), fpath + file_type, fsize
            )

            self.importer.run()
            for message in self.importer.messages:
                self.log(message)

            if len(self.importer.messages) > 0:
                raise BadRequest("Problem loading file.")

            contract = Contract.get_dc_by_id(sess, self.contract_id)
            contract.update_state(state)
            sess.commit()
            self.log("Finished loading '" + fpath)
            return True

    def sftp_handler(self, sess, properties, contract):
        host_name = properties["hostname"]
        user_name = properties["username"]
        password = properties["password"]
        try:
            port = properties["port"]
        except KeyError:
            port = None
        file_type = properties["file_type"]
        directories = properties["directories"]
        state = contract.make_state()

        try:
            last_import_keys = state["last_import_keys"]
        except KeyError:
            last_import_keys = state["last_import_keys"] = {}

        sess.rollback()
        self.log("Connecting to sftp server at " + host_name + ":" + str(port) + ".")
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        ftp = pysftp.Connection(
            host_name, username=user_name, password=password, cnopts=cnopts
        )
        ftp.timeout = 120
        home_path = ftp.pwd

        f = None

        for directory in directories:
            self.log("Checking the directory '" + directory + "'.")
            try:
                last_import_key = last_import_keys[directory]
            except KeyError:
                last_import_key = last_import_keys[directory] = ""

            dir_path = home_path + "/" + directory
            ftp.cwd(dir_path)
            files = []
            for attr in ftp.listdir_attr():
                fpath = dir_path + "/" + attr.filename
                try:
                    ftp.cwd(fpath)
                    continue  # directory
                except paramiko.SFTPError:
                    pass

                key = str(attr.st_mtime) + "_" + attr.filename
                if key > last_import_key:
                    files.append((key, fpath))

            if len(files) > 0:
                f = sorted(files)[0]
                last_import_keys[directory] = f[0]
                break

        if f is None:
            self.log("No new files found.")
            ftp.close()
            self.log("Logged out.")
            return False
        else:
            key, fpath = f
            self.log("Attempting to download " + fpath + " with key " + key + ".")
            f = tempfile.TemporaryFile()
            ftp.getfo(fpath, f)
            self.log("File downloaded successfully.")
            ftp.close()
            self.log("Logged out.")

            f.seek(0, os.SEEK_END)
            fsize = f.tell()
            f.seek(0)
            self.log("File size is " + str(fsize) + " bytes.")
            self.log("Treating files as type " + file_type)
            self.importer = HhDataImportProcess(
                self.contract_id, 0, TextIOWrapper(f, "utf8"), fpath + file_type, fsize
            )

            self.importer.run()
            messages = self.importer.messages
            self.importer = None
            for message in messages:
                self.log(message)

            if len(messages) > 0:
                raise BadRequest("Problem loading file.")

            contract = Contract.get_dc_by_id(sess, self.contract_id)
            contract.update_state(state)
            sess.commit()
            self.log("Finished loading '" + fpath)
            return True


def https_handler(sess, log_f, properties, contract, now=None):
    url_template_str = properties["url_template"]
    url_values = properties.get("url_values", {})
    download_days = properties["download_days"]
    result_data_key = properties["result_data_key"]
    if now is None:
        now = utc_datetime_now()
    window_finish = utc_datetime(now.year, now.month, now.day) - HH
    window_start = utc_datetime(now.year, now.month, now.day) - Timedelta(
        days=download_days
    )
    log_f(f"Window start: {hh_format(window_start)}")
    log_f(f"Window finish: {hh_format(window_finish)}")
    env = jinja2.Environment(autoescape=True, undefined=jinja2.StrictUndefined)
    url_template = env.from_string(url_template_str)
    for era in (
        sess.query(Era)
        .filter(
            Era.dc_contract == contract,
            Era.start_date <= window_finish,
            or_(Era.finish_date == null(), Era.finish_date >= window_start),
        )
        .distinct()
    ):
        chunk_start = hh_max(era.start_date, window_start)
        chunk_finish = hh_min(era.finish_date, window_finish)
        for mpan_core in (era.imp_mpan_core, era.exp_mpan_core):
            if mpan_core is None:
                continue

            log_f(f"Looking at MPAN core {mpan_core}.")

            vals = {"chunk_start": chunk_start, "chunk_finish": chunk_finish}
            vals.update(url_values.get(mpan_core, {}))
            try:
                url = url_template.render(vals)
            except jinja2.exceptions.UndefinedError as e:
                raise BadRequest(
                    f"Problem rendering the URL template: {url_template_str}. "
                    f"The problem is: {e}. This can be fixed by editing the "
                    f"properties of this contract."
                )

            log_f(f"Retrieving data from {url}.")

            sess.rollback()  # Avoid long transactions
            res = requests.get(url, timeout=120)
            res.raise_for_status()
            result = res.json()
            if isinstance(result, dict):
                result_data = result[result_data_key]
            elif isinstance(result, list):
                result_data = result
            else:
                raise BadRequest(
                    f"Expecting a JSON object at the top level, but instead got "
                    f"{result}"
                )
            raw_data = []
            for jdatum in result_data:
                raw_data.append(
                    dict(
                        mpan_core=mpan_core,
                        start_date=utc_datetime(1, 1, 1)
                        + Timedelta(seconds=jdatum["Time"] / 10000000),
                        channel_type="ACTIVE",
                        value=jdatum["Value"],
                        status="A",
                    )
                )
            HhDatum.insert(sess, raw_data, contract)
            sess.commit()
    log_f("Finished loading.")
    return False


def get_hh_import_task(contract):
    return tasks.get(contract.id)


def startup_contract(contract_id):
    task = HhImportTask(contract_id)
    tasks[contract_id] = task
    task.start()


def startup():
    for procs in processes.values():
        for proc in procs:
            if proc.is_alive():
                raise BadRequest(
                    "Can't start hh importer, there are still some hh imports running."
                )

    with Session() as sess:
        for contract in (
            sess.query(Contract)
            .join(MarketRole)
            .filter(MarketRole.code == "C")
            .order_by(Contract.id)
        ):
            startup_contract(contract.id)


@atexit.register
def shutdown():
    for task in tasks.values():
        task.stop()

    for task in tasks.values():
        task.join()

    for procs in processes.values():
        for proc in procs:
            proc.join()
